unit GcObj;
// GC objects
//
// Version 1.0b
// Copyright 1997 Vladimir A. Merzlaikov (wanderer@npt.rnd.runnet.ru)
//
// Memory for all objects allocated from GC memory
// All object, which have TGcObject as parent - uncollectable objects
// TGcObject               - GC Object,
// TGcAtomicObject         - GC Object without pointers
// TGCUncollectableObject  - Uncollectable object with poiters to GC Objects
// TGCStubbornctableObject - Freezable GC objects

interface

type
  // Garbage collectable (GC) object, which can point to GC objects
  TGcObject = class(TObject)
  private
    class function NewInstance: TObject; override;
    procedure FreeInstance; override;
  end;

  // GC object, without pointers to GC objects
  TGcAtomicObject = class(TGcObject)
  private
    class function NewInstance: TObject; override;
  end;

  // Non GC object, which can point to GC objects
  TGcUncollectableObject = class(TGcObject)
  private
    class function NewInstance: TObject; override;
  end;

  // Non GC object, which can't point to GC objects
  TGcAtomicUncollectableObject = class(TGcObject)
  private
    class function NewInstance: TObject; override;
  end;

  // Garbage collectable (GC) object, which can point to GC objects
  // and change only between BeginChange and  EndChage calls
  // or after creating and before first EndChange
  TGcStubbornObject = class(TGcObject)
  private
    fCanChange : Boolean;
    procedure SetCanChange( aCan : Boolean );
  public
    property CanChange : Boolean read fCanChange write SetCanChange;
    procedure BeginChange;
    procedure EndChange;
  private
    class function NewInstance: TObject; override;
  end;

implementation
uses Gc,GcMM;

procedure InvokeFinalizer( aObj : Pointer; aData : Pointer); far; cdecl;
begin
  TGCObject(aObj).Destroy;
end;

//== TGcObject ============================================
class function TGcObject.NewInstance;
var aOldProc : TGC_finalization_proc;
    aOldData : Pointer;
begin
  Result:=InitInstance(GC_malloc(InstanceSize));
  if not Assigned(Result) then exit;
  GC_register_finalizer_ignore_self(Result,InvokeFinalizer,Result,aOldProc,aOldData);
end;

procedure TGcObject.FreeInstance;
var aOldProc : TGC_finalization_proc;
    aOldData : Pointer;
begin
  GC_register_finalizer_ignore_self(Self,nil,aOldData,aOldProc,aOldData);
  GC_free(Self);
end;

//== TGcAtomicObject ======================================
class function TGcAtomicObject.NewInstance;
var aOldProc : TGC_finalization_proc;
    aOldData : Pointer;
begin
  Result:=InitInstance(GC_malloc(InstanceSize));
  if not Assigned(Result) then exit;
  GC_register_finalizer(Result,InvokeFinalizer,Result,aOldProc,aOldData);
end;

//== TGcUncollectableObject ===============================
class function TGcUncollectableObject.NewInstance;
var aOldProc : TGC_finalization_proc;
    aOldData : Pointer;
begin
  Result:= Self.InitInstance(GC_malloc_uncollectable(InstanceSize));
  if not Assigned(Result) then exit;
  GC_register_finalizer(Result,InvokeFinalizer,Result,aOldProc,aOldData);
end;

//== TGcAtomicUncollectableObject =============================
class function TGcAtomicUncollectableObject.NewInstance;
var aOldProc : TGC_finalization_proc;
    aOldData : Pointer;
begin
  Result:= Self.InitInstance(GC_malloc_atomic_uncollectable(InstanceSize));
  if not Assigned(Result) then exit;
  GC_register_finalizer(Result,InvokeFinalizer,Result,aOldProc,aOldData);
end;

//== TGcStubbornObject ====================================
class function TGcStubbornObject.NewInstance;
begin
  Result:= Self.InitInstance(GC_malloc_stubborn(InstanceSize));
  TGcStubbornObject(Result).fCanChange:=true;
end;

procedure TGcStubbornObject.SetCanChange;
begin
  if fCanChange=aCan then exit;
  fCanChange:=aCan;
  if aCan
    then GC_change_stubborn(Self)
    else GC_end_stubborn_change(Self)
  {end};
end;

procedure TGcStubbornObject.BeginChange;
begin
  fCanChange:=true;
  GC_change_stubborn(Self)
end;

procedure TGcStubbornObject.EndChange;
begin
  fCanChange:=false;
  GC_end_stubborn_change(Self)
end;

initialization
finalization
  GC_gcollect; // Collect garbage object - run its destructors
end.
